RoosterTools = LibStub("AceAddon-3.0"):NewAddon("RoosterTools", "AceConsole-3.0", "AceEvent-3.0", "AceTimer-3.0",
                   "AceComm-3.0", "AceSerializer-3.0")
RoosterTools.version = GetAddOnMetadata("Rooster_Tools", "Version")
RoosterTools.debug = false

local db = nil
local dbconn = nil 
local util = RoosterToolsUtil

function RoosterTools:CharacterReview(guid)
    local c = db.characters[guid]

    if (not c) then
        return
    end

    local now = time()
    if (now - c.updated > 1440) then
        c.isAFK = nil
        c.isConnected = nil
        c.isVisible = nil
        c.withinInspectDistance = nil
        c.withinTradeDistance = nil
        c.withinDuelDistance = nil
        c.withinFollowDistance = nil
        c.partyLeader = nil
        c.partyAssist = nil
        c.isDead = nil
        c.unitIndex = nil
    end
end

function RoosterTools:GroupCleanup(guid)
    local g = db.groups[guid]

    if (not g) then
        return
    end

    local now = time()
    if (now - g.updated > 1440 and util.count(g.instances) == 0) then
        db.groups[guid] = nil
    end
end

function RoosterTools:GroupReview(group)
    local guilds = {}

    local findJoin = function(events)
        for k, v in pairs(events) do
            if (v.state == "JOIN_GROUP") then
                return v
            end
        end
    end

    local t = {}
    table.insert(t, group)
    for k, v in pairs(group.members) do
        local events = RoosterTools.GroupEventStream(k, t)
        local joinedEvent = findJoin(events)

        local member = db.characters[k]
        v.unitName = member.unitName
        v.guildName = member.guildName

        if (joinedEvent and member.guildName) then
            if (not guilds[member.guildName]) then
                guilds[member.guildName] = {}
            end

            table.insert(guilds[member.guildName], joinedEvent.time)
        end
    end

    -- Calulate when this group became a guild group and for which guild.
    local guildGroupTime = 9007199254740994
    local guildName = nil
    local guildMembers = nil
    for k, b in pairs(guilds) do
        local count = util.count(b)
        if (count >= db.config[group.type].minGuildMembers) then
            table.sort(b)
            local groupTime = b[db.config[group.type].minGuildMembers]
            guildGroupTime = math.min(guildGroupTime, groupTime)
            guildName = k
            guildMembers = count
        end
    end

    -- Update the group with a guild group time and guild name.
    if (guildName) then
        group.guildName = guildName
        group.guildGroupTime = guildGroupTime
        group.guildMembers = guildMembers
    end
end

-- Sort chronologically.
function RoosterTools.SortEvents(events)
    if (not events) then
        return
    end

    local sortRank = {
        JOIN_GROUP = 0,
        ACTIVE = 1,
        INSTANCE_START = 2,
        ENCOUNTER_START = 3,
        ENCOUNTER_END = 4,
        INSTANCE_END = 5,
        GROUP_END = 6,
    }

    table.sort(events, function(a, b)
        if (a.time == b.time) then
            if ((sortRank[a.state] or 0) < (sortRank[b.state] or 0)) then
                return true
            else 
                return false
            end
        elseif (a.time < b.time) then
            return true
        else
            return false
        end
    end)
end

function RoosterTools.GroupEventStream(playerGuid, groups)
    local allEvents = {}

    -- Find all events that the player was a part of...
    for k, v in pairs(groups) do
        local member = v.members[playerGuid]
        if (member) then
            local events = {}

            -- Grab the states of the player to compare with the encounters.
            for k, v in pairs(member.states) do
                table.insert(events, {
                    state = v.state,
                    time = v.time
                })
            end

            RoosterTools.SortEvents(events)

            table.insert(events, {
                state = "JOIN_GROUP",
                time = events[1]["time"] or 0
            })

            -- Get all the possible encounters and add them to the timeline.
            for k, v in pairs(v.encounters) do
                table.insert(events, {
                    state = "ENCOUNTER_START",
                    time = v.started,
                    encounterName = v.encounterName
                })
                table.insert(events, {
                    state = "ENCOUNTER_END",
                    time = v.ended,
                    encounterName = v.encounterName
                })
            end

            table.insert(events, {
                state = "GROUP_END", 
                time = v.updated, 
                instance = v.currentInstance, 
                created = v.created 
            })

            for k, v in pairs(v.instances) do
                table.insert(events, {
                    state = "INSTANCE_START",
                    type = v.type,
                    time = v.created,
                    instance = v.instanceName,
                    created = v.created
                })
                table.insert(events, {
                    state = "INSTANCE_END",
                    time = v.updated,
                    type = v.type,
                    instance = v.instanceName,
                    created = v.created
                })
            end

            RoosterTools.SortEvents(events)
            for k, v in pairs(events) do
                table.insert(allEvents, v)
            end
        end
    end
    
    return allEvents
end

function RoosterTools:Calc(playerName)
    -- Find the character record
    local playerGuid, player = dbconn:SelectCharacterByName(playerName)
    if (not player) then
        return nil
    end

    local results = {}

    local lastState = PartyWatch.ROOSTER_STATES.INACTIVE
    local lastEventTime = 0
    local activeTime = 0
    local xActiveTime = 0
    local inactiveTime = 0
    local guildGroups = dbconn.SelectGuildGroups()
    local events = self.GroupEventStream(playerGuid, guildGroups)

    local buff = ""
    for k, v in pairs(events) do
        buff = buff .. k .. " ".. v.state .. " " ..v.time .. " " 
        if (v.type) then buff = buff .. v.type end
        buff = buff .. "\n"

        if (lastEventTime == 0) then
            lastEventTime = v.time
        end
        
        if (v.state == PartyWatch.ROOSTER_STATES.ACTIVE) then
            activeTime = activeTime + difftime(v.time, lastEventTime)
            lastState = v.state
        elseif (v.state == PartyWatch.ROOSTER_STATES.XACTIVE) then
            xActiveTime = xActiveTime + difftime(v.time, lastEventTime)
            lastState = v.state
        elseif (v.state == PartyWatch.ROOSTER_STATES.INACTIVE) then
            inactiveTime = inactiveTime + difftime(v.time, lastEventTime)
            lastState = v.state
        elseif (lastState ~= PartyWatch.ROOSTER_STATES.INACTIVE) then
            if (lastState == PartyWatch.ROOSTER_STATES.XACTIVE) then
                xActiveTime = xActiveTime + difftime(v.time, lastEventTime)
            elseif (lastState == PartyWatch.ROOSTER_STATES.ACTIVE) then
                activeTime = activeTime + difftime(v.time, lastEventTime)
            end
        end

        lastEventTime = v.time

        if (v.state == "GROUP_END") then
            lastState = PartyWatch.ROOSTER_STATES.INACTIVE
            lastEventTime = 0
            activeTime = 0
            xActiveTime = 0
            inactiveTime = 0
        elseif (v.state == "INSTANCE_END") then
            table.insert(results, {
                ["created"] = v.created,
                ["instance"] = v.instance,
                ["instanceType"] = v.type,
                ["activeTime"] = activeTime,
                ["xActiveTime"] = xActiveTime,
                ["inactiveTime"] = inactiveTime,
                ["guid"] = v.guid
            })

            lastState = PartyWatch.ROOSTER_STATES.INACTIVE
            lastEventTime = 0
            activeTime = 0
            xActiveTime = 0
            inactiveTime = 0
        end
    end
    db.temp = buff
    return results
end

function RoosterTools:GuildRosterTableUpdate()
    local guildName = GetGuildInfo("player");

    local groups = dbconn:SelectGuildRaidGroups()
    local totalRaids = util.count(groups)

    local numTotalMembers, numOnlineMaxLevelMembers, numOnlineMembers = GetNumGuildMembers();

    local updateGuild = function(unitName, guildName)
        for playerGuid, player in pairs(db.characters) do
            if (player.unitName and player.unitName == unitName) then
                player.guildName = guildName
            end
        end
    end

    local mains = {}
    local roster = {}
    for i = 1, numTotalMembers do
        local name, rank, rankIndex, level, class, zone, note, officernote, online, status, classFileName,
            achievementPoints, achievementRank, isMobile, isSoREligible, standingID = GetGuildRosterInfo(i);

        if (name) then
            local nameOnly = name:gmatch("[^-]+")()

            -- Determine this toon's main.
            local main = db.mains[nameOnly]
            if (main and main.name) then
                main = main.name
            else
                main = nameOnly
            end

            if (not mains[main]) then
                mains[main] = {
                    raids = {},
                    score = 0,
                    other = 0,
                    raidPercent = 0
                }
            end

            if (main == nameOnly) then
                mains[main].rankIndex = rankIndex
                mains[main].rank = rank
                mains[main].name = name
            end

            -- Update toons we know about. 
            updateGuild(nameOnly, RoosterTools.player.guildName)

            local results = RoosterTools:Calc(nameOnly)

            local score = 0
            local raids = 0.0
            local other = 0
            if (results) then
                for k, v in pairs(results) do
                    if (v.instanceType == "raid") then
                        -- Total the raid only one time per main. 
                        mains[main].raids[k] = 1
                        
                        local s = v.activeTime + v.xActiveTime
                        mains[main].score = mains[main].score + s

                        -- Round the score to the nearest point per 15 minutes
                        local mult = 10^0
                        s = s / 900
                        s =  math.floor(s * mult + 0.5) / mult
                        
                        score = score + s
                        raids = raids + 1
                    else
                        mains[main].other = mains[main].other + 1
                        other = other + 1
                    end
                end
            end

            local raidPercent = 0
            if (totalRaids > 0) then
                local mainRaids = util.count(mains[main].raids)
                mains[main].raidPercent = mainRaids * 100 / totalRaids
                mains[main].absent = totalRaids - mainRaids
                raidPercent = raids * 100 / totalRaids
            end

            local row = {rankIndex, rank, main, name, other, raids, totalRaids - raids, raidPercent .. "%", score}
            table.insert(roster, row)
        end
    end

    local aliasIt = function(tbl) 
        for k, v in pairs(tbl) do
            local alias = db.aliases[v[3]]
            if (alias and alias.alias) then
                v[3] = alias.alias
            end
        end

        return tbl
    end

    -- Set the total guild mains on the frame.
    RoosterTools.rootFrame.guildMembers:SetText(util.count(mains))

    if (RoosterTools.rootFrame.mainsOnly) then
        local mainRoster = {}
        for k, v in pairs(mains) do
            local s = v.score
            -- Round the score to the nearest point per 15 minutes
            local mult = 10^0
            s = s / 900
            s =  math.floor(s * mult + 0.5) / mult
            local row = {v.rankIndex, v.rank, k, v.name, v.other, util.count(v.raids), v.absent, v.raidPercent .. "%", s}
            table.insert(mainRoster, row)
        end

        self.guildRosterTable:SetData(aliasIt(mainRoster), true)
    else
        self.guildRosterTable:SetData(aliasIt(roster), true)
    end
end

function RoosterTools:OnInitialize()
    util.logInfo("version: " .. RoosterTools.version)
    
    RoosterToolsComm:OnInitialize()
    RoosterToolsChat:OnInitialize()
    RoosterToolsEvent:OnInitialize()
    RoosterToolsDBConn:OnInitialize()

    self.db = RoosterToolsDB
    db = RoosterToolsDB
    self.dbconn = RoosterToolsDBConn
    dbconn = RoosterToolsDBConn
    self.util = util

    -- Clear out old state that can get stuck if the user logs out before leaving a party
    if (not IsInGroup()) then
        db.partyGuid = nil
    end
    
    -- Store standard info about the player.
    local guildName = GetGuildInfo("player");
    local guid = UnitGUID("player")
    local playerName = UnitName("player")
    self.player = {
        ["guid"] = guid,
        ["guildName"] = guildName,
        ["playerName"] = playerName
    }

    -- Clean up old character records.
    for k, v in pairs(db.characters) do
        self:CharacterReview(k)
    end

    for k, v in pairs(db.groups) do
        self:GroupCleanup(k)
        self:GroupReview(v)
    end

    -- Every "interval" execute the update.
    local interval = 15 -- seconds
    self.npsTimer = self:ScheduleRepeatingTimer(function()
        -- Update our current data with the latest from our current party.
        PartyWatch:UpdateDB()
    end, interval)
end
